package VDRTools::File;

use strict;
use Cwd;
use vars qw(
	@ISA
	@EXPORT
	$ERRORS
	$MAX_DEPTH
	$BUFFER_SIZE
);

require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
	vt_cp
	vt_mv
	vt_rm
	vt_mkdir
	vt_absolute
	vt_filename
	vt_path
	vt_size
	vt_list_dirfiles
);

use VDRTools qw(error);

$ERRORS = {
	NOARGS		=> 'No arguments passed to function',
	NOFILES		=> 'No files passed to function',
	BADOPTS		=> 'Unknown option %s passed to function',
	NOCREATEDIR	=> 'Can not create directory "%s"',
	NOCREATEDIR	=> 'Can not chmod "%s"',
	NODIR		=> '%s is not a directory',
	NOOPENDIR	=> 'Can not open directory "%s"',
	READOPEN	=> 'Can not open file "%s" for reading',
	WRITEOPEN	=> 'Can not open file "%s" for writing',
	NOREAD		=> 'Can not read file "%s"',
	NOWRITE		=> 'Can not write file "%s"',
	NOCHMOD		=> 'Can not chmod file "%s"',
	NOCHOWN		=> 'Can not chown file "%s"',
	NORMFILE	=> 'Can not remove file "%s"',
	NORMDIR		=> 'Can not remove directory "%s"',
};

$MAX_DEPTH = 1000;
$BUFFER_SIZE = 1024 * 1024 * 2;

######################################################################

sub vt_mv {
	my ( $from, $to, $opts ) = @_;
	
	$opts = {} unless ( $opts and ref $opts eq 'HASH' );
	
	unless ( $from ) {
		error( $ERRORS->{NOARGS}, 'VDRTools::File::vt_mv()' );
		return;
	}
	
	unless ( $to ) {
		error( 'No target defined', 'VDRTools::File::vt_mv()' );
		return;
	}
	
	unless ( -e $from ) {
		error( 'Source file "' . $from . '" not found', 'VDRTools::File::vt_mv()' );
		return;
	}
	
	my $owner = delete( $opts->{owner} );
	$owner = getpwuid( $> ) unless ( defined( $owner ) );
	my ( $uid, $gid );
	if ( defined( $owner ) and $owner ne 'root' ) {
		if ( getpwuid( $> ) ne 'root' ) {
			error( 'You must be root to use the owner option', 'VDRTools::File::vt_mv()' );
			return;
		}
		
		my @owner = getpwnam( $owner );
		
		unless ( @owner ) {
			error( 'No such user "' . $owner . '" passed with owner option', 'VDRTools::File::vt_mv()' );
			return;
		}
		
		$uid = $owner[2];
		$gid = $owner[3];
	}
	
	if ( keys( %$opts ) ) {
		error( sprintf( $ERRORS->{BADOPTS}, ( join ', ', keys( %$opts ) ) ), 'VDRTools::File::vt_mv()' );
		return;
	}
	
	$from = vt_absolute( $from );
	$to = vt_absolute( $to );
	
	if ( vt_cp( $from, $to, { owner => $owner } ) ) {
		return unless ( vt_rm( $from ) );
	}
	else {
		error( 'Can not move "' . $from . '" to "' . $to . '"', 'VDRTools::File::vt_mv()' );
		return;
	}
	
	return 1;
}

sub vt_rm {
	my ( @files ) = @_;
	
	unless ( @files ) {
		error( $ERRORS->{NOARGS}, 'VDRTools::File::vt_rm()' );
		return;
	}
	
	foreach my $file ( @files ) {
		$file = vt_absolute( $file );
		next unless ( -e $file );
		if ( -l $file ) {
			unless ( unlink( $file ) ) {
				error( sprintf( $ERRORS->{NORMFILE}, $file ), 'VDRTools::File::vt_rm()' );
				return;
			}
		}
		elsif ( -d $file ) {
			return unless ( vt_rm( vt_list_dirfiles( $file, { follow_symlinks => 0 } ) ) );
			unless ( rmdir( $file ) ) {
				error( sprintf( $ERRORS->{NORMDIR}, $file ), 'VDRTools::File::vt_rm()' );
				return;
			}
		}
		else {
			unless ( unlink( $file ) ) {
				error( sprintf( $ERRORS->{NORMFILE}, $file ), 'VDRTools::File::vt_rm()' );
				return;
			}
		}
	}
	return 1;
}

sub vt_cp {
	my ( $from, $to, $opts ) = @_;
	
	$opts = {} unless ( $opts and ref $opts eq 'HASH' );
	
	unless ( $from ) {
		error( $ERRORS->{NOARGS}, 'VDRTools::File::vt_cp()' );
		return;
	}
	
	unless ( $to ) {
		error( 'No target defined', 'VDRTools::File::vt_cp()' );
		return;
	}
	
	unless ( -e $from ) {
		error( 'Source file "' . $from . '" not found', 'VDRTools::File::vt_cp()' );
		return;
	}
	
	my $owner = delete( $opts->{owner} );
	$owner = getpwuid( $> ) unless ( defined( $owner ) );
	my ( $uid, $gid );
	if ( defined( $owner ) and $owner ne 'root' ) {
		if ( getpwuid( $> ) ne 'root' ) {
			error( 'You must be root to use the owner option', 'VDRTools::File::vt_cp()' );
			return;
		}
		
		my @owner = getpwnam( $owner );
		
		unless ( @owner ) {
			error( 'No such user "' . $owner . '" passed with owner option', 'VDRTools::File::vt_cp()' );
			return;
		}
		
		$uid = $owner[2];
		$gid = $owner[3];
	}
	
	if ( keys( %$opts ) ) {
		error( sprintf( $ERRORS->{BADOPTS}, ( join ', ', keys( %$opts ) ) ), 'VDRTools::File::vt_cp()' );
		return;
	}
	
	$from = vt_absolute( $from );
	$to = vt_absolute( $to );
	
	if ( -d $to ) {
		my $newfile = vt_filename( $from );
		my ( $perms ) = ( lstat( $from ) )[2];
		if ( -l $from ) {
			symlink( readlink( $from ), $to . '/' . $newfile );
		}
		if ( -d $from ) {
			return unless ( vt_mkdir( $to . '/' . $newfile, { perms => $perms, owner => $owner } ) );
			foreach ( vt_list_dirfiles( $from, { follow_symlinks => 0 } ) ) {
				return unless ( vt_cp( $_, $to . '/' . $newfile, { owner => $owner } ) );
			}
		}
		else {
			return unless ( vt_cp( $from, $to . '/' . $newfile, { owner => $owner } ) );
		}
	}
	else {
		my $newfile = vt_filename( $to );
		( $to ) = vt_path( $to );
		unless ( -d $to ) {
			error( 'Target directory "' . $to . '" does not exist', 'VDRTools::File::vt_cp()' );
			return;
		}
		if ( -l $from ) {
			symlink( readlink( $from ), $to . '/' . $newfile );
		}
		elsif ( -d $from ) {
			error( 'Target directory "' . $to . '/' . $newfile . '" is a file', 'VDRTools::File::vt_cp()' );
			return;
		}
		else {
			my ( $perms ) = ( lstat( $from ) )[2];
			
			local ( *FROM, *TO );
			
			unless ( open( FROM, "< $from" ) ) {
				error( sprintf( $ERRORS->{READOPEN}, $from ), 'VDRTools::File::vt_cp()' );
				return;
			}
			unless ( open( TO, ">" . $to . '/' . $newfile ) ) {
				error( sprintf( $ERRORS->{WRITEOPEN}, $to . '/' . $newfile ), 'VDRTools::File::vt_cp()' );
				return;
			}
			
			binmode( FROM );
			binmode( TO );
			
			my $buf_size = -s FROM;
			$buf_size = $BUFFER_SIZE if ( $buf_size > $BUFFER_SIZE );
			$buf_size = 1024 if ($buf_size < 512);
			
			while ( 1 ) {
				my ($r, $w, $t, $buf);
				$r = sysread( FROM, $buf, $buf_size );
				unless ( defined( $r ) ) {
					error( sprintf( $ERRORS->{NOREAD}, $from ), 'VDRTools::File::vt_cp()' );
					return;
				}
				last unless $r;
				for ($w = 0; $w < $r; $w += $t) {
					$t = syswrite( TO, $buf, $r - $w, $w);
					unless ( defined( $t ) ) {
						error( sprintf( $ERRORS->{NOWRITE}, $to . '/' . $newfile ), 'VDRTools::File::vt_cp()' );
						return;
					}
				}
			}
			
			close( FROM );
			close( TO );
			
			if ( $uid and $gid ) {
				unless ( chown( $uid, $gid, $to . '/' . $newfile ) ) {
					error( sprintf( $ERRORS->{NOCHOWN}, $to . '/' . $newfile ), 'VDRTools::File::vt_cp()' );
					return;
				}
			}
			
			unless ( chmod( $perms, $to . '/' . $newfile ) ) {
				error( sprintf( $ERRORS->{NOCHMOD}, $to . '/' . $newfile ), 'VDRTools::File::vt_cp()' );
				return;
			}
		}
	}
	return 1;
}

sub vt_size {
	my ( $file, $opts ) = @_;
	
	$opts = {} unless ( $opts and ref $opts eq 'HASH' );
	
	unless ( $file ) {
		error( $ERRORS->{NOARGS}, 'VDRTools::File::vt_size()' );
		return;
	}
	
	$file = vt_absolute( $file );
	
	my $format = delete( $opts->{format} );
	$format = 'B' unless ( defined( $format ) );
	
	if ( keys( %$opts ) ) {
		error( sprintf( $ERRORS->{BADOPTS}, ( join ', ', keys( %$opts ) ) ), 'VDRTools::File::vt_size()' );
		return;
	}
	
	my $size = -s $file;
	
	if ( -d $file ) {
		map( $size += -s $_ , vt_list_dirfiles( $file, { recursive => 1, follow_symlinks => 0 } ) );
	}
	
	if ( $format eq 'K' ) {
		$size /= 1024;
	}
	elsif ( $format eq 'M' ) {
		$size /= ( 1024 * 1024 );
	}
	if ( $format eq 'G' ) {
		$size /= ( 1024 * 1024 * 1024 );
	}
	$size = int( $size );
	
	return $size;
}

sub vt_list_dirfiles {
	my ( $dir, $opts ) = @_;
	
	$opts = {} unless ( $opts and ref $opts eq 'HASH' );
	
	unless ( $dir ) {
		error( $ERRORS->{NOARGS}, 'VDRTools::File::vt_list_dirfiles()' );
		return;
	}
	
	$dir = vt_absolute( $dir );
	
	unless ( -d $dir ) {
		error( sprintf( $ERRORS->{NODIR}, $dir ), 'VDRTools::File::vt_list_dirfiles()' );
		return;
	}
	
	my $only_files = delete( $opts->{only_files} );
	$only_files = 0 unless ( defined( $only_files ) );
	
	my $only_dirs = delete( $opts->{only_dirs} );
	$only_dirs = 0 unless ( defined( $only_dirs ) );
	
	my $regexp = delete( $opts->{regexp} );
	$regexp = '.*' unless ( defined( $regexp ) );
	
	my $recursive = delete( $opts->{recursive} );
	$recursive = 0 unless ( defined( $recursive ) );
	
	my $max_depth = delete( $opts->{max_depth} );
	$max_depth = $MAX_DEPTH unless ( defined( $max_depth ) );
	
	my $follow_symlinks = delete( $opts->{follow_symlinks} );
	$follow_symlinks = 1 unless ( defined( $follow_symlinks ) );
	
	if ( keys( %$opts ) ) {
		error( sprintf( $ERRORS->{BADOPTS}, ( join ', ', keys( %$opts ) ) ), 'VDRTools::File::vt_list_dirfiles()' );
		return;
	}
	
	local *DIR;
	unless ( opendir( DIR, $dir ) ) {
		error( sprintf( $ERRORS->{NOOPENDIR}, $dir ), 'VDRTools::File::vt_list_dirfiles()' );
		return;
	}
	my @files = readdir( DIR );
	closedir( DIR );
	
	my @list;
	
	foreach my $file ( @files ) {
		next if ( $file eq '.' or $file eq '..' );
		my $entry = ( ( $dir eq '/' ) ? '' : $dir ) . '/' . $file;
		if ( -d $entry and $recursive and $max_depth > 0 ) {
			my $test_sym = 1;
			
			if ( -l $entry ) {
				$test_sym = 0 if ( not $follow_symlinks );
				my $target = readlink( $entry );
				my ( $t_path, $t_isrelative ) = vt_path( $target );
				if ( $t_isrelative ) {
					my $t_filename = vt_filename( $target );
					( $target ) = vt_path( $dir . '/' . $t_path . '/' . $t_filename );
					$target .= '/' . $t_filename;
				}
				$test_sym = 0 if ( $entry =~ /$target/ );
			}
			
			if ( $test_sym ) {
				push( @list, vt_list_dirfiles( $entry,
						{
							only_files	=> $only_files,
							only_dirs	=> $only_dirs,
							regexp		=> $regexp,
							recursive	=> $recursive,
							max_depth	=> $max_depth - 1,
							follow_symlinks	=> $follow_symlinks,
						}
					)
				);
			}
		}
		if ( $regexp =~ /\\\// ) {
			next if ( $entry !~ /$regexp/ );
		}
		else {
			next if ( $file !~ /$regexp/ );
		}
		next if ( $only_files and -d $entry );
		next if ( $only_dirs and not -d $entry );
		
		push( @list, $entry );
	}
	
	return @list;
}


sub vt_mkdir {
	my ( $dir, $opts ) = @_;
	
	$opts = {} unless ( $opts and ref $opts eq 'HASH' );
	
	unless ( $dir ) {
		error( $ERRORS->{NOARGS}, 'VDRTools::File::vt_mkdir()' );
		return;
	}
	
	$dir = vt_absolute( $dir );
	
	my $perms = delete( $opts->{perms} );
	$perms = 0755 unless ( defined( $perms ) );
	
	my $owner = delete( $opts->{owner} );
	my ( $uid, $gid );
	if ( defined( $owner ) and $owner ne 'root' ) {
		if ( getpwuid( $> ) ne 'root' ) {
			error( 'You must be root to use the owner option', 'VDRTools::File::vt_mkdir()' );
			return;
		}
		
		my @owner = getpwnam( $owner );
		
		unless ( @owner ) {
			error( 'No such user "' . $owner . '" passed with owner option', 'VDRTools::File::vt_mkdir()' );
			return;
		}
		
		$uid = $owner[2];
		$gid = $owner[3];
	}
	
	if ( keys( %$opts ) ) {
		error( sprintf( $ERRORS->{BADOPTS}, ( join ', ', keys( %$opts ) ) ), 'VDRTools::File::vt_mkdir()' );
		return;
	}
	
	if ( -d $dir ) {
		error( 'Directory "' . $dir . '" already exists', 'VDRTools::File::vt_mkdir()' );
		return;
	}
	
	my @path = split( /\//, $dir );
	
	my $new_path;
	
	foreach ( @path ) {
		next if ( not length( $_ ) );
		$new_path .= '/' . $_;
		unless ( -d $new_path ) {
			unless ( mkdir( $new_path, $perms ) ) {
				error( sprintf( $ERRORS->{NOCREATEDIR}, $new_path ), 'VDRTools::File::vt_mkdir()' );
				return;
			}
			
			if ( $uid and $gid ) {
				unless ( chown( $uid, $gid, $new_path ) ) {
					error( sprintf( $ERRORS->{NOCHOWN}, $new_path ), 'VDRTools::File::vt_mkdir()' );
					return;
				}
			}
		}
	}
	
	return 1;
}

sub vt_absolute {
	my ( @files ) = @_;
	
	unless ( @files ) {
		error( $ERRORS->{NOFILES}, 'VDRTools::File::vt_absolute()' );
		return;
	}
	
	my $cwd;
	
	my @out_files;
	
	foreach ( @files ) {
		my ( $path, $isrelative ) = vt_path( $_ );
		if ( $isrelative ) {
			$cwd ||= cwd() || getcwd();
			unless ( $cwd ) {
				error( 'Could not get cwd', 'VDRTools::File::vt_absolute()' );
				return;
			}
			( $path ) = vt_path( $cwd . '/' . $_ );
		}
		my $filename = vt_filename( $_ );
		
		if ( $filename eq '.' ) {
			push( @out_files, $path );
		}
		elsif ( $filename eq '..' ) {
			push( @out_files, @{vt_path( $path )}[0] );
		}
		else {
			if ( $path eq '/' ) {
				push( @out_files, $path . $filename );
			}
			else {
				push( @out_files, $path . '/' . $filename );
			}
		}
	}
	
	if ( length( @out_files ) == 1 ) {
		return $out_files[0];
	}
	else {
		return @out_files;
	}
}

sub vt_filename {
	my $file = shift;
	
	unless ( $file ) {
		error( $ERRORS->{NOARGS}, 'VDRTools::File::vt_filename()' );
		return;
	}
	
	$file =~ s/\/$//;
	
	my @path = split( /\//, $file );
	
	my $filename = pop( @path );
	
	return $filename;
}

sub vt_path {
	my $file = shift;
	
	unless ( $file ) {
		error( $ERRORS->{NOARGS}, 'VDRTools::File::vt_path()' );
		return;
	}
	
	$file =~ s/\/$//g;
	
	return ( '', 0 ) unless ( $file );
	
	my @path = split( /\//, $file );
	
	pop( @path );
	
	my $isrelative = ( $file =~ /^\// ) ? 0 : 1;
	
	my $i = 0;
	
	my @new_path;
	
	foreach ( @path ) {
		if ( $_ eq '.' or not length( $_ ) ) {
			next;
		}
		elsif ( $_ eq '..' ) {
			$i--;
		}
		else {
			$i++;
		}
		if ( not $isrelative and $i < 0 and $_ eq '..' ) {
			next;
		}
		elsif ( $i < 0 and $_ eq '..' ) {
			push( @new_path, ".." );
		}
		elsif ( $i >= 0 and $_ eq '..' ) {
			pop( @new_path );
		}
		else {
			push( @new_path, $_ );
		}
	}
	
	my $new_path = join "/", @new_path;
	$new_path = $isrelative ? "./$new_path" : "/$new_path";
	
	return ( $new_path, $isrelative );
}

1;

__END__

=pod

=head1 NAME

VDRTools::File

=head1 FUNCTIONS

=head2 vt_cp()

=over 4

use VDRTools::File;

vt_cp( $file1, $file2, $opts );

$file1 and $file2 can be both a file or a directory. If $file2 is a
directory $file1 will be copied into $file2.
Returns 1 on success and 0 if an error occurs.
$opts is a hashref of options. Possible options are:

=item owner

Changes the owner of the copied files. You must be root to use this
option.

=back

=head2 vt_mv()

=over 4

use VDRTools::File;

vt_mv( $file1, $file2, $opts );

$file1 and $file2 can be both a file or a directory. If $file2 is a
directory $file1 will be moved into $file2.
Returns 1 on success and 0 if an error occurs.
$opts is a hashref of options. Possible options are:

=item owner

Changes the owner of the moved files. You must be root to use this
option.

=back

=head2 vt_rm()

=over 4

use VDRTools::File;

vt_rm( $file );

$file can be both a file or a directory. If $file is a directory, it will
be removed recursively.
Returns 1 on success and 0 if an error occurs.

=back

=head2 vt_mkdir()

=over 4

use VDRTools::File;

vt_mkdir( $dir, $opts );

Creates the directory $dir, recursively if necessary.
Returns 1 on success and 0 if an error occurs.
$opts is a hashref of options. Possible options are:

=item owner

Sets the owner of the created directory. You must be root to use this
option.

=item perms

Sets the permissions of the created directory. Default is 0755.

=back

=head2 vt_absolute()

=over 4

use VDRTools::File;

$abs = vt_absolute( $rel );

Returns the absolute path of a file.

=back

=head2 vt_filename()

=over 4

use VDRTools::File;

$filename = vt_filename( $filename_with_path );

Returns the file portion of a filename.

=back

=head2 vt_path()

=over 4

use VDRTools::File;

( $path, $isrelative ) = vt_path( $filename_with_path );

Returns the directory portion of a filename and if it is relative or not.

=back

=head2 vt_size()

=over 4

use VDRTools::File;

$size = vt_size( $filename, $opts )

Returns the size of a file or directory.
$opts is a hashref of options. Possible options are:

=item format

Defines the output format. Possible values are 'B', 'K', 'M' or 'G'.
Default is 'B'.

=back

=head2 vt_list_dirfiles()

=over 4

use VDRTools::File;

@files = vt_list_dirfiles( $dir, $opts );

Returns a list of files of the given directory.
$opts is a hashref of options. Possible options are:

=item only_dirs

Possible values are 0 or 1. Default is 0.

=item only_files

Possible values are 0 or 1. Default is 0.

=item recursive

Possible values are 0 or 1. Default is 0.

=item max_depth

Default is 1000.

=item follow_symlinks

Possible values are 0 or 1. Default is 1.

=item regexp

Default is '.*'.

=cut

