package VDRTools::Recording;

use strict;
use vars qw(
	@ISA
	@EXPORT
	$ERRORS
	$BUFFER_SIZE
);

require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
	vt_create_index
	vt_join_rec
	vt_update_videodir
	vt_HMSF_to_frames
	vt_frames_to_HMSF
);

use VDRTools qw(error);
use VDRTools::File;
use VDRTools::System;
use VDRTools::Network qw(vt_svdrpsend);

$ERRORS = {
	NOARGS		=> 'No arguments passed to function',
	NODIR		=> '%s is not a directory',
	NOVIDEOFILES	=> 'No video files found in "%s"',
	READOPEN	=> 'Can not open file "%s" for reading',
	WRITEOPEN	=> 'Can not open file "%s" for writing',
	NOREAD		=> 'Can not read file "%s"',
	NOWRITE		=> 'Can not write file "%s"',
	BADOPTS		=> 'Unknown option %s passed to function',
};

$BUFFER_SIZE = 1024 * 1024 * 2;

######################################################################

sub vt_update_videodir {
	utime( time(), time(), $VDRTools::VIDEODIR . '/.update' );
}

sub vt_join_rec {
	my ( $recs, $opts ) = @_;
	
	$opts = {} unless ( $opts and ref $opts eq 'HASH' );
	
	unless ( $recs ) {
		error( $ERRORS->{NOARGS}, 'VDRTools::Recording::vt_join_rec()' );
		return;
	}
	
	unless ( ref $recs eq 'ARRAY' ) {
		error( 'The first argument must be a arrayref', 'VDRTools::Recording::vt_join_rec()' );
		return;
	}
	
	my $svdrpsend = delete( $opts->{svdrpsend} );
	$svdrpsend = 0 unless ( defined( $svdrpsend ) );
	
	if ( scalar( @$recs ) < 2 ) {
		error( 'You must pass at least two recordings to function', 'VDRTools::Recording::vt_join_rec()' );
		return;
	}
	
	my $rec_arrayref;
	
	foreach my $rec ( @$recs ) {
		my $tmp_hash;
		$tmp_hash->{dir} = vt_absolute( $rec );
		
		unless ( -d $tmp_hash->{dir} ) {
			error( sprintf( $ERRORS->{NODIR}, $tmp_hash->{dir} ), 'VDRTools::Recording::vt_join_rec()' );
			return;
		}
		
		push( @{$tmp_hash->{video_files}}, vt_list_dirfiles( $tmp_hash->{dir}, { only_files => 1, regexp => '^\d{3}\.vdr$' } ) );
		
		unless ( scalar( @{$tmp_hash->{video_files}} ) ) {
			error( sprintf( $ERRORS->{NOVIDEOFILES}, $tmp_hash->{dir} ), 'VDRTools::Recording::vt_join_rec()' );
			return;
		}
		
		unless ( -e $tmp_hash->{dir} . '/index.vdr' ) {
			return unless ( vt_create_index( $tmp_hash->{dir} ) );
		}
		
		push( @$rec_arrayref, $tmp_hash );
	}
	
	my $videodir = delete( $opts->{videodir} );
	$videodir = $VDRTools::VIDEODIR unless ( defined( $videodir ) );
	
	unless ( -d $videodir ) {
		error( sprintf( $ERRORS->{NODIR}, $videodir ), 'VDRTools::Recording::vt_join_rec()' );
		return;
	}
	
	my $priority = delete( $opts->{priority} );
	$priority = vt_vdr_setup_parm( 'DefaultPriority' ) unless ( defined( $priority ) or $priority =~ /^\d\d$/ );
	
	my $lifetime = delete( $opts->{lifetime} );
	$lifetime = vt_vdr_setup_parm( 'DefaultLifetime' ) unless ( defined( $lifetime ) or $lifetime =~ /^\d\d$/ );
	
	my $maxfilesize = delete( $opts->{maxfilesize} );
	$maxfilesize = vt_vdr_setup_parm( 'MaxVideoFileSize' ) unless ( defined( $maxfilesize ) or $maxfilesize =~ /^\d+$/ );
	
	$maxfilesize *= ( 1024 * 1024 );
	
	my $newrec;
	
	my $name = delete( $opts->{name} );
	unless ( defined( $name ) ) {
		( $newrec ) = vt_path( $rec_arrayref->[0]->{dir} );
		$newrec .= '_new';
		$newrec .= '/' . vt_filename( $rec_arrayref->[0]->{dir} );
	}
	else {
		$newrec = $videodir . '/' . $name;
	
		my ( undef, $min, $hour, $day, $month, $year ) = localtime( time() );
		
		$newrec .= '/' . ( $year + 1900 ) . '-' . ( ( ( $month += 1 ) < 10 ) ? '0' : '' ) . $month . '-' . ( ( $day < 10 ) ? '0' : '' ) . $day
			. '.' . ( ( $hour < 10 ) ? '0' : '' ) . $hour . '.' . ( ( $min < 10 ) ? '0' : '' ) . $min . '.'
			. $priority . '.' . $lifetime . '.rec';
	}
	
	if ( keys( %$opts ) ) {
		error( sprintf( $ERRORS->{BADOPTS}, ( join ', ', keys( %$opts ) ) ), 'VDRTools::Recording::vt_join_rec()' );
		return;
	}
	
	if ( -d $newrec ) {
		error( 'Directory ' . $newrec . ' already exists', 'VDRTools::Recording::vt_join_rec()' );
		return;
	}
	
	my ( $testdir ) = vt_path( $newrec );
	
	while ( not -d $testdir ) {
		( $testdir ) = vt_path( $testdir );
	}
	
	my ( $device, undef, undef, $avail ) = vt_df( $testdir );
	
	my $total_size = 0;
	map( $total_size += vt_size( $_->{dir} ), @$rec_arrayref );
	
	if ( $total_size > $avail ) {
		error( 'Not enough disc space left on device ' . $device, 'VDRTools::Recording::vt_join_rec()' );
		vt_svdrpsend( 'MESG "Not enough disc space left on device ' . $device . '"' ) if ( $svdrpsend );
		return;
	}
	
	return unless ( vt_mkdir( $newrec ) );
	
	vt_cp( $rec_arrayref->[0]->{dir} . '/summary.vdr', $newrec . '/' ) if ( -e $rec_arrayref->[0]->{dir} . '/summary.vdr' );
	vt_cp( $rec_arrayref->[0]->{dir} . '/info.vdr', $newrec . '/' ) if ( -e $rec_arrayref->[0]->{dir} . '/info.vdr' );
	
	foreach my $rec_hash ( @$rec_arrayref ) {
		my $filetowrite = _vt_get_video_file_to_write( $newrec, $maxfilesize );
		
		my ( $addsize, $bytesread );
		
		if ( -e $newrec . '/' . $filetowrite ) {
			$addsize = $maxfilesize - vt_size( $newrec . '/' . $filetowrite );
		}
		else {
			$addsize = $maxfilesize;
		}
		
		local ( *MARKS, *INDEX, *NEWINDEX, *READ, *WRITE );
		
		my @marks;
		
		if ( -e $rec_hash->{dir} . '/marks.vdr' ) {
			unless ( open ( MARKS, $rec_hash->{dir} . '/marks.vdr' ) ) {
				error( sprintf( $ERRORS->{READOPEN}, $rec_hash->{dir} . '/marks.vdr' ), 'VDRTools::Recording::vt_join_rec()' );
				return;
			}
			while(<MARKS>) {
				if ( $_ =~ /^(\d:\d\d:\d\d\.\d\d)/ ) {
					push( @marks, vt_HMSF_to_frames( $1 ) );
				}
			}
			close( MARKS );
		}
		
		my @char;
		
		unless ( open( INDEX, $rec_hash->{dir} . '/index.vdr' ) ) {
			error( sprintf( $ERRORS->{READOPEN}, $rec_hash->{dir} . '/index.vdr' ), 'VDRTools::Recording::vt_join_rec()' );
			return;
		}
		for ( my $i = 0; $i < 8; $i++ ) {
			$char[$i] = getc( INDEX );
		}
		
		my $offset = hex( sprintf( "%02X%02X%02X%02X", ord($char[3]), ord($char[2]), ord($char[1]), ord($char[0]) ) );
		my $filetoread = sprintf( "%03d.vdr", hex( sprintf( "%02X", ord($char[5]) ) ) );
		
		seek( INDEX, 0, 0 );
		unless ( open( READ, $rec_hash->{dir} . '/' . $filetoread ) ) {
			error( sprintf( $ERRORS->{READOPEN}, $rec_hash->{dir} . $filetoread ), 'VDRTools::Recording::vt_join_rec()' );
			return;
		}
		unless ( open( WRITE, ">>" . $newrec . '/' . $filetowrite ) ) {
			error( sprintf( $ERRORS->{WRITEOPEN}, $rec_hash->{dir} . $filetoread ), 'VDRTools::Recording::vt_join_rec()' );
			return;
		}
		unless ( open( NEWINDEX, ">>" . $newrec . '/index.vdr' ) ) {
			error( sprintf( $ERRORS->{WRITEOPEN}, $newrec . '/index.vdr' ), 'VDRTools::Recording::vt_join_rec()' );
			return;
		}
		
		my $eof = 0;
		
		while ( not eof( INDEX ) ) {
			for(my $i = 0; $i < 16; $i++){
				if ( eof( INDEX ) ) {
					$eof = 1;
					last;
				}
				$char[$i] = getc( INDEX );
			}
			
			foreach( @marks ) {
				if( $_ == tell( INDEX ) / 8 - 2) {
					unless ( open( NEWMARKS, ">>" . $newrec . '/marks.vdr' ) ) {
						error( sprintf( $ERRORS->{WRITEOPEN}, $newrec . '/marks.vdr' ), 'VDRTools::Recording::vt_join_rec()' );
						return;
					}
					printf( NEWMARKS "%s\n", vt_frames_to_HMSF( tell( NEWINDEX ) / 8 ) );
					close( NEWMARKS );
				}
			}
			
			my $pos = tell( WRITE );
			my ( $fileno ) = $filetowrite =~ /(\d\d\d)\.vdr/;
			my $indexframe = chr( hex( substr( sprintf( "%08X", $pos ), 6, 2 ) ) )
				. chr( hex( substr( sprintf( "%08X", $pos ), 4, 2) ) )
				. chr( hex( substr( sprintf( "%08X", $pos ), 2, 2) ) )
				. chr( hex( substr( sprintf( "%08X", $pos ), 0, 2) ) )
				. $char[4] . chr( $fileno ) . $char[6] . $char[7];
			print NEWINDEX $indexframe;
			
			my $frame;
			
			if ( $eof ) {
				seek( READ, $offset, 0 );
				read( READ, $frame , vt_size( $rec_hash->{dir} . '/' . $filetoread ) - $offset );
				print WRITE $frame;
			}
			
			else {
				seek( INDEX, -8 , 1 );
				my $nextoffset = hex( sprintf( "%02X%02X%02X%02X", ord($char[11]), ord($char[10]), ord($char[9]), ord($char[8]) ) );
				if ( $nextoffset < $offset ) {
					seek( READ, $offset, 0 );
					$bytesread += read( READ, $frame , vt_size( $rec_hash->{dir} . '/' . $filetoread ) - $offset );
					print WRITE $frame;
					close( READ );
					$filetoread = sprintf( "%03d.vdr", hex( sprintf( "%02X", ord($char[13]) ) ) );
					open( READ, $rec_hash->{dir} . '/' . $filetoread );
				}
				else {
					seek( READ, $offset, 0 );
					$bytesread += read( READ, $frame , $nextoffset - $offset );
					print WRITE $frame;
				}
				$offset = $nextoffset;
				
				if (
					( $bytesread >= $addsize ) and
					(
						(
							( hex( sprintf( "%02X", ord($char[12]) ) ) == 1 ) and
							( hex( sprintf( "%02X", ord($char[5]) ) ) == hex( sprintf( "%02X", ord($char[13]) ) ) )
						) or
						( hex( sprintf( "%02X", ord($char[5]) ) ) < hex( sprintf( "%02X", ord($char[13]) ) ) )
					)
				) {
					close( WRITE );
					$filetowrite = _vt_get_video_file_to_write( $newrec, $maxfilesize );
					unless ( open( WRITE, ">>" . $newrec . '/' . $filetowrite ) ) {
						error( sprintf( $ERRORS->{WRITEOPEN}, $newrec . '/' . $filetowrite ), 'VDRTools::Recording::vt_join_rec()' );
						return;
					}
					$addsize = $maxfilesize;
					$bytesread = 0;
				}
			}
			
		}
		
		close( IN );
		close( OUT );
		close( INDEX );
		close( NEWINDEX );
	}
	
	return 1;
}

sub _vt_get_video_file_to_write {
	my $dir = shift;
	my $maxfilesize = shift || ( vt_vdr_setup_parm( 'MaxVideoFileSize' ) * 1024 * 1024 );
	my $filetowrite;
	my @todirfiles = vt_list_dirfiles( $dir, { only_files => 1, regexp => '^\d{3}\.vdr$' } );
	
	@todirfiles = sort( map( vt_filename( $_ ), @todirfiles ) );
	
	if ( $#todirfiles >= 0 ) {
		if ( vt_size( $dir . '/' . $todirfiles[-1] ) >= $maxfilesize ) {
			my ( $oldno ) = $todirfiles[-1] =~ /(\d{3})\.vdr/;
			$filetowrite = sprintf("%03d.vdr", $oldno + 1);
		}
		else {
			$filetowrite = $todirfiles[-1];
		}
	}
	else {
		$filetowrite = "001.vdr";
	}
	
	return $filetowrite;
}

####
#### vt_HMSF_to_frames and vt_frames_to_HMSF adapted from vdr/recording.c
####

sub vt_HMSF_to_frames {
	my $HMSF = shift;
	$HMSF =~ /^(\d):(\d\d):(\d\d)\.(\d\d)/;
	return ($1 * 3600 + $2 * 60 + $3) * 25 + $4 - 1;
}

sub vt_frames_to_HMSF {
	my $frames = shift;
	my ($H, $M, $S, $F);
	$F = ($frames % 25) + 1;
	$S = ($frames / 25);
	$M = $S / 60 % 60;
	$H = $S / 3600;
	$S %= 60;
	return sprintf("%d:%02d:%02d.%02d", $H, $M, $S, $F);
}

##
## vt_create_index is from Dr. Peter Sebbel's alias doc 
## (peter@vdr-portal.de) create_indexvdr.pl
##

sub vt_create_index {
	my $dir = shift;
	
	unless ( $dir ) {
		error( $ERRORS->{NOARGS}, 'VDRTools::Recording::vt_create_index()' );
		return;
	}
	
	$dir = vt_absolute( $dir );
	
	unless ( -d $dir ) {
		error( sprintf( $ERRORS->{NODIR}, $dir ), 'VDRTools::Recording::vt_create_index()' );
		return;
	}
	
	my $genindex = vt_find_prog( 'genindex' );
	
	if ( $genindex ) {
		if ( -e $dir . '/index.vdr.generated' ) {
			return 1 if ( vt_mv( $dir . '/index.vdr.generated', $dir . '/index.vdr' ) );
		}
		return 1 if ( -e $dir . '/index.vdr' );
	}
	
	my @vdr_files = vt_list_dirfiles( $dir, { only_files => 1, regexp => '^\d{3}\.vdr$' } );
	
	unless ( @vdr_files ) {
		error( sprintf( $ERRORS->{NOVIDEOFILES}, $dir ), 'VDRTools::Recording::vt_create_index()' );
		return;
	}
	
	if ( -e $dir . '/index.vdr' ) {
		return unless ( vt_mv( $dir . '/index.vdr', $dir . '/oldindex.vdr' ) );
	}
	
	local *INDEX;
	unless ( open( INDEX, ">" . $dir . '/index.vdr' ) ) {
		error( sprintf( $ERRORS->{WRITEOPEN}, $dir . '/index.vdr' ), 'VDRTools::Recording::vt_create_index()' );
		return;
	}
	
	foreach my $file ( sort( @vdr_files ) ) {
		my ( $fileno ) = vt_filename( $file ) =~ /(\d{3})\.vdr/;
		my $offset = 0;
		my $file_offset = 0;
		my $last = 0;
		my %pes_header;
		my $buffer;
		my $ret;
		my $picstart = pack ("H8", "00000100");
		my $pacstart = pack ("H6", "000001");
		
		local *IN;
		
		unless ( open( IN, "< $file") ) {
			error( sprintf( $ERRORS->{READOPEN}, $file ), 'VDRTools::Recording::vt_create_index()' );
			return;
		}
		
		binmode( IN );
		
		my $buf_size = $BUFFER_SIZE;
		
		while ( not $last or $offset < length( $buffer ) ) {
			my $tmp_buffer;
			my $head_tmpl = 'H6 H2 n1 B8 B8 C1 B40';
			
			if ( not $last and ( not defined( $buffer ) or ( $offset + $pes_header{PES_packet_length} + 150 ) > ( length( $buffer ) ) ) ) {
				
				$ret = sysread( IN, $tmp_buffer, $buf_size );
		
				unless ( defined( $ret ) ) {
					error( sprintf( $ERRORS->{NOREAD}, $file ), 'VDRTools::Recording::vt_create_index()' );
					return;
				}
				
				$buffer = substr( $buffer, $offset );
				
				$buffer .= $tmp_buffer;
				
				$last = 1 if ( length( $buffer ) < $buf_size );
				
				$offset = 0;
			}
			
			(
				$pes_header{startcode},
				$pes_header{stream_id},
				$pes_header{PES_packet_length},
				$pes_header{eflags1},
				$pes_header{eflags2},
				$pes_header{header_length},
				$pes_header{PTS_raw_bits}
			) = unpack ( $head_tmpl, substr($buffer, $offset, 14) );
			
			if ( $pes_header{startcode} ne "000001" ) {
				my $tmpoffset = index( $buffer, $pacstart, $offset );
				if ( $tmpoffset == -1 ) {
					$file_offset += ( length($buffer) - $offset );
					$offset = length( $buffer );
				}
				else {
					$file_offset += $tmpoffset - $offset;
					$offset = $tmpoffset;
				}
				next;
			}
			
			my $decimal_code = hex($pes_header{stream_id});
			
			if ( ( $decimal_code <= 191 ) and ( $decimal_code < 240 ) and ( $decimal_code != 189) ) {
				$offset +=  $pes_header{PES_packet_length} + 6;
				$file_offset +=  $pes_header{PES_packet_length} + 6;
				next;
			}
			
			$pes_header{payload_start} = (9 + $pes_header{header_length});
			
			my $packet = substr( $buffer, $offset, $pes_header{PES_packet_length} + 6 );
			my $payload = substr( $packet, $pes_header{payload_start} );
			
			if ( ( $pes_header{stream_id} =~ /^e/ ) && ( $payload =~ m/$picstart(.{4})/ )   ) {
				my $pictype = "0b0" . substr( unpack( "B16", $1 ), 10, 3 );
				my $bin_code = pack( "ICCS", $file_offset, oct( $pictype ), $fileno , 0 );
				print INDEX $bin_code;
			}
			
			$offset +=  $pes_header{PES_packet_length} + 6;
			$file_offset +=  $pes_header{PES_packet_length} + 6;
		}
		
		close( IN );
	}
	
	close( INDEX );
	
	return 1;
}

1;

__END__

