package VDRTools::System;

use strict;
use vars qw(
	@ISA
	@EXPORT
	$ERRORS
);

require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
	vt_filesystems
	vt_df
	vt_find_prog
	vt_vdr_setup_parm
	vt_pidof
	vt_nice
	vt_fork_to_bg
);

use VDRTools qw(error);

$ERRORS = {
	READOPEN	=> 'Can not open file "%s" for reading',
	NOARGS		=> 'No arguments passed to function',
	PIPEFROM	=> 'Can not open pipe from "%s"',
	BADOPTS		=> 'Unknown option %s passed to function',
};

######################################################################

sub vt_fork_to_bg {
	my $pid = fork();
	if($pid != 0) {
		exit(0);
	}
}

sub vt_nice {
	my $nice = shift || 19;
	setpriority ( 0, 0, $nice );
}

sub vt_pidof {
	my $prog = shift || return;
	my $cmd = 'pidof ' . $prog;
	my $pid = qx($cmd);
	chomp( $pid );
	return unless ( $pid );
	return $pid;
}

sub vt_vdr_setup_parm {
	my $parm = shift || return;
	my $value;
	
	local *SETUP;
	unless ( open( SETUP, $VDRTools::VDRCONF ) ) {
		error( sprintf( $ERRORS->{READOPEN}, $VDRTools::VDRCONF ), 'VDRTools::System::vt_filesystems()' );
		return;
	}
	while( <SETUP> ) {
		if ( /^\Q$parm\E = (.*)/ ){
			$value = $1;
			chomp( $value );
			last;
		}
	}
	close( SETUP );
	
	return $value;
}

sub vt_find_prog {
	my $prog = shift;
	
	unless ( $prog ) {
		error( $ERRORS->{NOARGS}, 'VDRTools::System::vt_find_prog()' );
		return;
	}
	
	my $prog = VDRTools::File::vt_filename( $prog );
	
	foreach my $path ( split( /\:/, $ENV{PATH} ) ) {
		return $path . '/' . $prog if ( -x $path . '/' . $prog );
	}
	
	return undef;
}

sub vt_df {
	my ( $file, $opts ) = @_;
	
	$opts = {} unless ( $opts and ref $opts eq 'HASH' );
	
	unless ( $file ) {
		error( $ERRORS->{NOARGS}, 'VDRTools::System::vt_df()' );
		return;
	}
	
	my $format = delete( $opts->{format} );
	$format = 'B' unless ( defined( $format ) );
	
	if ( keys( %$opts ) ) {
		error( sprintf( $ERRORS->{BADOPTS}, ( join ', ', keys( %$opts ) ) ), 'VDRTools::System::vt_size()' );
		return;
	}
	
	$file = VDRTools::File::vt_absolute( $file );
	
	my $cmd = 'df ' . $file;
	
	local *CMD;
	unless ( open( CMD, "$cmd|" ) ) {
		error( sprintf( $ERRORS->{PIPEFROM}, $cmd ), 'VDRTools::System::vt_df()' );
		return;
	}
	my @ret = <CMD>;
	close( CMD );
	
	my ( $device, $total, $used, $avail, $percent, $mount );
	
	foreach ( @ret ) {
		chomp( $_ );
		next if ( $_ =~ /^(Filesystem|Dateisystem)/ );
		( $device, $total, $used, $avail, $percent, $mount ) = split( /\s+/, $_ );
		last;
	}
	
	unless ( $device ) {
		error( '', 'System::vt_df()' );
		return undef;
	}
	
	if ( $format eq 'B' ) {
		$total *= 1024;
		$used *= 1024;
		$avail *= 1024;
	}
	elsif ( $format eq 'M' ) {
		$total /= 1024;
		$used /= 1024;
		$avail /= 1024;
	}
	if ( $format eq 'G' ) {
		$total /= ( 1024 * 1024 );
		$used /= ( 1024 * 1024 );
		$avail /= ( 1024 * 1024 );
	}
	$total = int( $total );
	$used = int( $used );
	$avail = int( $avail );
	
	return ( $device, $total, $used, $avail, $percent, $mount );
}

# This funktion is from Sys::Filesystem
sub vt_filesystems {
	my $fstab = '/etc/fstab';
	my $mtab = '/etc/mtab';
	
	my $filesystems;

	# Default fstab and mtab layout
	my @keys = qw(fs_spec fs_file fs_vfstype fs_mntops fs_freq fs_passno);

	local *FSTAB;
	if ( open( FSTAB, "< $fstab" ) ) {
		while (<FSTAB>) {
			next if (/^\s*#/ || /^\s*$/);
			my @vals = split(/\s+/, $_);
			if ($vals[0] =~ /^\s*LABEL=(.+)\s*$/) {
				$filesystems->{$vals[1]}->{label} = $1;
			}
			$filesystems->{$vals[1]}->{mount_point} = $vals[1];
			$filesystems->{$vals[1]}->{device} = $vals[0];
			$filesystems->{$vals[1]}->{unmounted} = 1;
			$filesystems->{$vals[1]}->{special} = 1 if grep(/^$vals[2]$/,qw(swap proc devpts tmpfs usbdevfs));
			for (my $i = 0; $i < @keys; $i++) {
				$filesystems->{$vals[1]}->{$keys[$i]} = $vals[$i];
			}
		}
		close( FSTAB );
	}
	else {
		error( sprintf( $ERRORS->{READOPEN}, $fstab ), 'VDRTools::System::vt_filesystems()' );
		return;
	}

	# Read the mtab
	local *MTAB;
	if ( open( MTAB, "< $mtab" ) ) {
		while ( <MTAB> ) {
			next if /^\s*\#/;
			next if /^\s*$/;
			my @vals = split(/\s+/, $_);
			delete $filesystems->{$vals[1]}->{unmounted} if exists $filesystems->{$vals[1]}->{unmounted};
			$filesystems->{$vals[1]}->{mounted} = 1;
			$filesystems->{$vals[1]}->{mount_point} = $vals[1];
			$filesystems->{$vals[1]}->{device} = $vals[0];
			$filesystems->{$vals[1]}->{special} = 1 if grep(/^$vals[2]$/,qw(swap proc devpts tmpfs usbdevfs));
			for (my $i = 0; $i < @keys; $i++) {
				$filesystems->{$vals[1]}->{$keys[$i]} = $vals[$i];
			}
		}
		close( MTAB );
	}
	else {
		error( sprintf( $ERRORS->{READOPEN}, $mtab ), 'VDRTools::System::vt_filesystems()' );
		return;
	}

	return $filesystems;
}

1;